#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "AppController.h"
#include "FileRing.h"
#include "ImageWindow.h"
#include "SelectionCtrl.h"
#include "PHDefs.h"

#define IS_CLOSE(a, b, tol) (fabs((a) - (b)) < tol)

@interface AppController (PrivateMethods)

- (void) setCheckMarkInMenu: (NSMenu *) menu forValue: (float) value;
- (void) addToMenu: (NSMenu *) menu fromDescr: (NSString *[][3]) descr
            amount: (int) amount selector: (SEL) sel;

@end // @interface AppController (PrivateMethods)

@implementation AppController

- (id) init
{
    
    [super init];

    _window = nil;
    _fileRing = [[FileRing alloc] initWithTypes: [NSImage imageFileTypes]];
    _selCtrl = [SelectionCtrl new];

    return self;
}

- (void) dealloc
{
    RELEASE(_selCtrl);
    RELEASE(_fileRing);
    RELEASE(_window);

    [super dealloc];
}

// {title, key equivalent, shape (y/x ratio)}
static NSString * shape_data[][3] =
{
    {@"No constraint", @"", @"0"},
    {@"Monitor (4:3)", @"", @"0.75"},
    {@"Wide (16:9)", @"", @"0.5625"},
    {@"Square (1:1)", @"", @"1"},
    {@"Monitor (3:4)", @"", @"1.3333333334"}
};

// {title, key equivalent, max. size after scale}
static NSString * scale_data[][3] =
{
    {@"Do not scale", @"", @"0"},
    //{@"Maximal standard size", @"", @"-1"},
    {@"2272", @"", @"2272"},
    {@"2048", @"", @"2048"},
    {@"1600", @"", @"1600"},
    {@"1280", @"", @"1280"},
    {@"1024", @"", @"1024"},
    {@"800", @"", @"800"},
    {@"640", @"", @"640"},
    {@"480", @"", @"480"},
    {@"320", @"", @"320"},
};

- (void) createMenu
{
    SEL dummy = @selector(dummy_method:);
    
    NSMenu * menu = AUTORELEASE( [NSMenu new] );

    [menu addItemWithTitle: @"Info Panel..."
          action: @selector(showInfoPanel:)
          keyEquivalent: @""];
    [menu addItemWithTitle: @"Selection"
          action: dummy
          keyEquivalent: @""];
    [menu addItemWithTitle: @"Open"
          action: @selector(open:)
          keyEquivalent: @"o"];
    [menu addItemWithTitle: @"Rotate CCW"
          action: @selector(rotateCCW:)
          keyEquivalent: @"r"];
    [menu addItemWithTitle: @"Rotate CW"
          action: @selector(rotateCW:)
          keyEquivalent: @"R"];
    [menu addItemWithTitle: @"Next"
          action: @selector(nextImage:)
          keyEquivalent: @"n"];
    [menu addItemWithTitle: @"Previous"
          action: @selector(prevImage:)
          keyEquivalent: @"p"];
    [menu addItemWithTitle: @"Clip"
          action: @selector(clip:)
          keyEquivalent: @"c"];
    [menu addItemWithTitle: @"Undo"
          action: @selector(undo:)
          keyEquivalent: @"z"];
    [menu addItemWithTitle: @"Save"
           action: @selector(save:)
           keyEquivalent: @"s"];
    [menu addItemWithTitle: @"Hide"
          action: @selector(hide:)
          keyEquivalent: @"h"];
    [menu addItemWithTitle:@"Quit"
          action:@selector(terminate:)
          keyEquivalent:@"q"];


    // "Shape" menu
    
    NSMenu * shape = AUTORELEASE( [NSMenu new] );

    [shape addItemWithTitle: @"Scale To" action: dummy keyEquivalent: @""];

    // Add shapes
    
    int num_shapes = sizeof(shape_data)/sizeof(*shape_data);
    [self addToMenu: shape fromDescr: shape_data amount: num_shapes
          selector: @selector(setSelectionShape:)];


    // Add Lossless JPEG item

    //[shape addItem: [NSMenuItem separatorItem]];
    [shape addItemWithTitle: @"Lossless JPEG"
           action: @selector(toggleLLJPEG:)
           keyEquivalent: @""];
           
    [self setCheckMarkInMenu: shape forValue: [_selCtrl shape] ];

    // Attach submenu 
    [menu setSubmenu: shape forItem: [menu itemWithTitle:@"Selection"]];
    
    // "Scale To" menu
    
    NSMenu * scale_to = AUTORELEASE( [NSMenu new] );

    int num_scales = sizeof(scale_data)/sizeof(*scale_data);
    [self addToMenu: scale_to fromDescr: scale_data amount: num_scales
          selector: @selector(setSelectionScaleTo:)];

    [self setCheckMarkInMenu: scale_to forValue: [_selCtrl span] ];

    // Attach submenu
    [shape setSubmenu: scale_to forItem: [shape itemWithTitle:@"Scale To"]];
    
    [NSApp setMainMenu: menu];
}

- (void) addToMenu: (NSMenu *) menu fromDescr: (NSString *[][3]) descr
            amount: (int) amount selector: (SEL) sel
{
    NSMenuItem * item;
    int i;
    for (i=0; i < amount; ++i)
    {
        item = [[NSMenuItem alloc] initWithTitle: descr[i][0]
                               action: sel
                               keyEquivalent: descr[i][1]];
        float value = [descr[i][2] floatValue];
        [item setRepresentedObject: [NSNumber numberWithFloat: value]];

        [menu addItem: item];
        RELEASE(item);
    }
}

- (void) createWindow
{
    _window = [[ImageWindow alloc] init];
}

- (void) loadKeyBindings
{
    NSBundle * bun = [NSBundle mainBundle];
    NSString * path = [bun pathForResource: @"KeyBindings" ofType: @"dict" ];

    if ( [[NSFileManager defaultManager] fileExistsAtPath: path] )
    {
        NSInputManager * inputManager = [NSInputManager currentInputManager];
        [inputManager loadBindingsFromFile: path];
    }
}

- (void) applicationWillFinishLaunching: (NSNotification *)not
{
    [self createMenu];
    [self createWindow];
    [self loadKeyBindings];

    // get first command-line argument (zeroth is the program name)
    NSArray * args = [[NSProcessInfo processInfo] arguments];


    if ([args count] > 1)
    {
        NSArray * ar1 =
            [args subarrayWithRange: NSMakeRange( 1, [args count]-1 )];

        [_fileRing insertPathArray: ar1 positioning: PosAtEnd];
        [self openFile: [_fileRing currentPath]];
    }
}

- (void) applicationDidFinishLaunching: (NSNotification *)not
{
    [_window center];
    [_window makeKeyAndOrderFront: nil];
}

- (void) applicationWillTerminate: (NSNotification *)not
{
    // TODO: remove after making proper cursor handling.
    
    // A workaround for cursor handling bug. Reset to default.
    [[NSCursor arrowCursor] set];
}


- (BOOL) applicationShouldTerminateAfterLastWindowClosed: (id)sender
{
    return YES;
}

- (void) open: (id)sender
{
    NSArray * image_types = [NSImage imageFileTypes];
    NSOpenPanel * panel = [NSOpenPanel openPanel];
    [panel setCanChooseDirectories: YES];
    [panel setAllowsMultipleSelection: NO];

    NSString * dir = (_fileRing) ? [_fileRing currentDir] : NSHomeDirectory();

    NSString * fileName = (_fileRing) ? [_fileRing currentName] : NULL;

    int result = [panel runModalForDirectory: dir
                        file: fileName
                        types: image_types ];
    if (result == NSOKButton)
    {
        NSString * selection = [[panel filenames] objectAtIndex: 0];
        if ( selection )
        {
            [_fileRing replaceWithDirContainingPath: selection];
            [self openFile: [_fileRing currentPath]];
        }
    }
}

- (void) openFile: (NSString *)fileName
{
    NSAssert(_window, @"open: '_window' is not initialized");
    if (fileName)
        [_window loadFile: fileName];
}

- (void) setRotationTo: (float) degrees
{
    NSAssert(_window, @"setRotationTo: '_window' is not initialized");

    [_window setRotationTo: degrees];
}

- (void) nextImage: (id)sender
{
    [self openFile: [_fileRing nextPath]];
}

- (void) prevImage: (id)sender
{
    [self openFile: [_fileRing prevPath]];
}

- (void) save: (id)sender
{
    NSAssert(_window, @"save: '_window' is not initialized");

    NSSavePanel * panel = [NSSavePanel savePanel];

    //[panel setRequiredFileType: @"tiff"];
    [panel setAllowedFileTypes:
               [NSArray arrayWithObjects: @"tiff", @"jpg", nil ]];
    
    NSString * dir = (_fileRing) ? [_fileRing currentDir] : NSHomeDirectory();

    NSString * fileName = (_fileRing) ? [_fileRing currentName] : NULL;

    if (NSOKButton == [panel runModalForDirectory: dir file: fileName])
    {
        NSString * fileToSave = [panel filename];
        if ( [_window saveToFile: fileToSave] )
        {
            // successfully saved
            [_fileRing insertPath: fileToSave withSiblings: NO
                       positioning: PosAfterCur];
            [self setRotationTo: 0];
            [self openFile: fileToSave];
        }
    }
    
}

- (void) setSelectionShape: (id)sender
{
    float shape = 0;
    id rep = [(NSMenuItem*)sender representedObject];
    if (rep)
        shape = [rep floatValue];

    [_selCtrl setShape: shape];

    [self setCheckMarkInMenu: [sender menu] forValue: shape];

}

- (void) setSelectionScaleTo: (id)sender
{
    float span = 0;
    id rep = [(NSMenuItem*)sender representedObject];
    if (rep)
        span = [rep floatValue];

    [_selCtrl setSpan: span];

    [self setCheckMarkInMenu: [sender menu] forValue: span];
}


- (void) setCheckMarkInMenu: (NSMenu *) menu forValue: (float) value
{
    Class number_class = [NSNumber class];

    NSEnumerator * items = [[menu itemArray] objectEnumerator];
    NSMenuItem * item;

    //float current_shape = [_selCtrl shape];
    const float TOL = 0.01;

    while ( (item = [items nextObject]) )
    {
        id rep_object = [item representedObject];
        if ([rep_object isKindOfClass: number_class])
        {
            int state = IS_CLOSE([rep_object floatValue], value, TOL) ?
                NSOnState : NSOffState;

            [item setState: state];
        }
    }
}

- (BOOL) validateMenuItem: (NSMenuItem *) item
{
  SEL action = [item action];
  if ( sel_eq(action, @selector(nextImage:)) ||
       sel_eq(action, @selector(prevImage:)) )
  {
    return ([_fileRing count] > 1) ? YES : NO;
  }

  if ( sel_eq(action, @selector(save:)) )
  {
    return [_fileRing currentPath] ? YES : NO;
  }

  return YES;
}

- (void) showInfoPanel: (id)sender
{
    //NSString * image_types = [[NSImage imageFileTypes] description];

    NSDictionary * options =
        [NSDictionary dictionaryWithObjectsAndKeys:
                      @"Simple image viewer and editor", @"ApplicationDescription",
                      @"http://www.vaisburd.net/PhotoClip/", @"URL",
                      [NSArray arrayWithObject:
                                   @"Tima Vaisburd <tima@vaisburd.net>"],
                      @"Authors",
                      @VERSION_STR, @"ApplicationVersion",
                      @"Copyright (C) 2003 - 2006 Tima Vaisburd", @"Copyright",
                      @"License: GNU LGPL", @"CopyrightDescription",
                      nil];
                      
                                               
    [NSApp orderFrontStandardInfoPanelWithOptions: options];
}

- (SelectionCtrl *) selectionCtrl
{
    return _selCtrl;
}

@end
