#include "ConstrainedRectCalculator.h"
#include <Foundation/NSException.h>
#include <math.h>
#include "LowPassFilter.h"
#include "MouseTracker.h"

#define SIGN(x) ((x) > 0 ? 1 : -1)

enum {SPAN_MIN, SPAN_MAX, SPAN_LINEAR, SPAN_QUADRATIC};
static int span_mode = SPAN_QUADRATIC;
//static int span_mode = SPAN_MAX;

static const float A_DEFAULT = 100;


@interface ConstrainedRectCalculator (PrivateMethods)

- (Interval) spanIntervalFromStart: (NSPoint) start end: (NSPoint) end
                             shape: (float) shape;
- (float) spanFromInterval: (Interval) dx;

@end // ConstrainedRectCalculator (PrivateMethods)


@implementation ConstrainedRectCalculator

- (void) dealloc
{
    [_filter release];
    [super dealloc];
}

- (id) initWithTracker: (MouseTracker *) tracker mouse: (NSPoint) pt
{
    // New Selection
    [super initWithTracker: tracker];
    _a = A_DEFAULT;
    _filter = [[LowPassFilter alloc] init];
    return self;
}

- (id) initWithTracker: (MouseTracker *) tracker mouse: (NSPoint) pt
                 start: (NSPoint) start end: (NSPoint) end
{
    // Move Corner
    [super initWithTracker: tracker];

    NSRect selRect = PHRectFromPoints(start, end);
    end = PHClosestPointOnRectEdge(pt, selRect);
    _mouse_to_end = PHMakeVect2D(end.x - pt.x, end.y - pt.y);

    Interval dx  = [self spanIntervalFromStart: start end: end
                         shape: [_tracker shape]];
    _a = MIN((dx.min - 1), A_DEFAULT);
    
    _filter = [[LowPassFilter alloc] initWithValue: dx.max];

    // NSLog(@"initWithTms: _a=%g", _a);
    
    return self;
}

- (NSPoint) correctEndPointFromStart: (NSPoint) start end: (NSPoint) end
{
    float dx1 = end.x - start.x;
    float dy1 = end.y - start.y;
    float shape = [_tracker shape];
    
    Interval intvl = [self spanIntervalFromStart:start end: end shape: shape];

    // increase _a if it's too small and minimal span increases.
    if (_a < A_DEFAULT && _a < intvl.min)
        _a = MIN(intvl.min, A_DEFAULT);
    
    float dx = [self spanFromInterval: intvl] * SIGN(dx1);
    float dy = fabsf(dx) * shape * SIGN(dy1);

    // NSLog(@"%@ a=%g dx=%.2f", PHStringFromInterval(dx_intvl), _a, dx);

    Vect2D bounded = PHBoundVect2D(start, PHMakeVect2D(dx, dy),
                                   [_tracker frame]);

    NSPoint result = PHMovePointByVect(start, bounded);
    return result;
}

- (Interval) spanIntervalFromStart: (NSPoint) start end: (NSPoint) end
                             shape: (float) shape
{
    NSAssert(shape > 0, @"shape <= 0");

    Interval result;
    float dx1 = end.x - start.x;
    float dx2 = (end.y - start.y) / shape;
    result.max = MAX(fabsf(dx1), fabsf(dx2));
    result.min = MIN(fabsf(dx1), fabsf(dx2));
    return result;
}

- (float) spanFromInterval: (Interval) dx
{

    NSAssert1(dx.min >= 0, @"dx.min=%f < 0", dx.min);
    NSAssert1(dx.max >= 0, @"dx.max=%f < 0", dx.max);
    NSAssert(dx.min <= dx.max, @"dx.min > dx.max");

    if (_a <= 0)
        return dx.max;

    float x = dx.max;
    float y = dx.min;
    float result = x;

    BOOL in_between = NO;
    switch (span_mode)
    {
        case SPAN_LINEAR:
            if (x <= _a)
                result = y;
            else if (y < _a && _a < x)
            {
                result =  x * y / _a;
                in_between = YES;
            }
            else // y >= a
                result = x;
            break;
        case SPAN_QUADRATIC:
            if (y <= _a)
            {
                result = x * y * (2 * _a - y) / (_a * _a);
                in_between = YES;
            }
            else
                result = x;
            break;
        case SPAN_MIN:
            result = y;
            break;
        case SPAN_MAX:
            result = x;
            break;
    }


    // running average

    if (in_between == NO)
    {
        // hold filter ready
        [_filter initWithValue: result];
    }
    else
    {
        // calculate running average and substitute result with it
        result = [_filter addNewValue: result];
    }

    return result;
}

@end // ConstrainedRectCalculator
