#include <Foundation/Foundation.h>
#include <AppKit/NSImage.h>
#include "FileRing.h"


NSString * PhMakePath(NSString * dir, NSString * path);
NSString * PhMakeAbsPath(NSString * path);

void PhPrintArray(NSArray * arr);

@interface FileRing ( PrivateMethods )

- (void) setWritePositionByMode: (int) posMode;
- (unsigned) internalInsertPath: (NSString *) path withSiblings: (BOOL) withS;
- (unsigned) insertAbsFilePath: (NSString *) path;
- (BOOL) isAllowedPath: (NSString *) path;

@end

@implementation FileRing

- (id) init
{
    _allowedTypes = nil;
    _files = [[NSMutableArray alloc] init];
    _pos = 0;
    _writePos = 0;
    return self;
}

- (void) dealloc
{
    [_allowedTypes release];
    [_files release];
    [super dealloc];
}

- (id) initWithTypes: (NSArray *) allowedTypes
{
    _allowedTypes =
        allowedTypes ? [[NSArray alloc] initWithArray: allowedTypes] : 0;

    _files = [[NSMutableArray alloc] init];
    _pos = 0;
    _writePos = 0;

    //PhPrintArray(_allowedTypes);
    
    return self;
}

- (id) initFromPathArray: (NSArray *)paths withTypes: (NSArray *) allowedTypes
{
    // Create allowed types
    _allowedTypes =
        allowedTypes ? [[NSArray alloc] initWithArray: allowedTypes] : 0;

    // Create files ring
    _files = [[NSMutableArray alloc] init];

    NSFileManager * fmgr = [NSFileManager defaultManager];
    NSString * currentDir = [fmgr currentDirectoryPath];
    
    NSString * path = nil;
    NSString * absPath = nil;

    _writePos = 0;
    _pos = 0;

    NSEnumerator * pathsIter = [paths objectEnumerator];
    while ( (path = [pathsIter nextObject]) != nil )
    {
        absPath = PhMakePath( currentDir, path );
        [self internalInsertPath: absPath withSiblings: NO];
    }    

    // PhPrintArray(_allowedTypes);
    // PhPrintArray(_files);
    // NSLog(@"    count=%d, pos=%d\n", [_files count], _pos);
    
    return self;
}

- (void) insertPathArray: (NSArray *) paths positioning: (int) posMode
{
    [self setWritePositionByMode: posMode];

    unsigned pos = NSNotFound;
    
    NSFileManager * fmgr = [NSFileManager defaultManager];
    NSString * currentDir = [fmgr currentDirectoryPath];
    
    NSString * path = nil;
    NSString * absPath = nil;

    NSEnumerator * pathsIter = [paths objectEnumerator];
    while ( (path = [pathsIter nextObject]) != nil )
    {
        absPath = PhMakePath( currentDir, path );
        unsigned pos1 = [self internalInsertPath: absPath withSiblings: NO];
        if ( pos == NSNotFound )
            pos = pos1;
    }

    if ( pos != NSNotFound)
        _pos = pos;
}

- (void) replaceWithDirContainingPath: (NSString *) path
{
    [self removeAllObjects];
    [self insertPath: path withSiblings: YES positioning: PosBegin];
}


// Insert path into the ring accoring to posMode. If path is a directory,
// expands that directory contents to one level down.

- (void) insertPath: (NSString *) path withSiblings: (BOOL) withSibl
        positioning: (int) posMode
{
    [self setWritePositionByMode: posMode];
    _pos = [self internalInsertPath: path withSiblings: withSibl];
}


// Inserts path or the directory contents at _writePos accordign to the
// following logic:
//
//  Path          withSibl   Action
//  ----------    --------   -----------
//  regular file   YES       Insert all allowed files from path's directory
//  regular file   NO        Insert one file
//  directory      YES       Insert the content of directory
//  directory      NO        The same (for directories withSibl is ignored)
//
// Returns the position of the first inserted or found element,
// or NSNotFound
//
- (unsigned) internalInsertPath: (NSString *) path withSiblings: (BOOL) withS
{
    path = PhMakeAbsPath(path);

    NSFileManager * fmgr = [NSFileManager defaultManager];

    BOOL isDir;
    unsigned pos = NSNotFound;

    if ([fmgr fileExistsAtPath:path isDirectory: &isDir] == NO)
    {
        NSLog(@"Skipping nonexisting path '%@'\n", path );
        return pos;
    }

    // Obtain the directory that contains the path.
    NSString * dir = (isDir) ? path : [path stringByDeletingLastPathComponent];

    if ( isDir == NO && withS == NO )
    {
        // Insert path at current write position
        if ( [self isAllowedPath: path])
            pos = [self insertAbsFilePath: path];

        return pos;
    }


    // Expand the directory into temporary matchedPaths array
    
    NSMutableArray * matchedPaths = [[NSMutableArray alloc] init];

    NSEnumerator * content
        = [[fmgr directoryContentsAtPath: dir] objectEnumerator];
    NSString * name = nil;
    NSString * absPath = nil;
    while ( (name = [content nextObject]) != nil )
    {
        absPath = PhMakePath(dir, name);
        NSDictionary * dict =
            [fmgr fileAttributesAtPath: absPath traverseLink: YES];
        
        if (dict && [[dict fileType] isEqualToString: NSFileTypeRegular])
            if ( [self isAllowedPath: absPath] )
                [matchedPaths addObject: absPath];
    }

    // Sort according to NSFilePanel representation
    [matchedPaths sortUsingSelector: @selector(caseInsensitiveCompare:)];

    // Add new elements from matchedPaths to the ring
    content = [matchedPaths objectEnumerator];
    while ( (absPath = [content nextObject]) != nil )
    {
        unsigned pos1 = [self insertAbsFilePath: absPath];
        if (pos == NSNotFound)
            pos = pos1;
    }

    // Override pos if the path was a regular file
    if ( isDir == NO ) // and withSibl == YES since we got here
        pos = [_files indexOfObject: PhMakeAbsPath( path )];

    return pos;
}


- (void) setWritePositionByMode: (int) posMode
{
    switch( posMode )
    {
        case PosBeforeCur:
            _writePos = _pos;
            break;
        case PosAfterCur:
            _writePos = _pos + 1;
            break;
        case PosBegin:
            _writePos = 0;
            break;
        case PosAtEnd:
            _writePos = [_files count];
            break;
        default:
            NSAssert1( 0, @"setWritePoitionByMode: invalid mode=%d",
                       posMode);
    }
}

- (unsigned) insertAbsFilePath: (NSString *) path
{
    NSAssert1( [path isAbsolutePath] == YES, @"Path %@ is not absolute", path);

    // Check whether path is already in the ring
    int pos = [_files indexOfObject: path ];
    if (pos != NSNotFound)
        return pos;

    // NSLog(@"FileRing: adding %@, pos=%d\n", path, _writePos);

    [_files insertObject: path atIndex: _writePos ];
    return _writePos++;
}

- (BOOL) isAllowedPath: (NSString *) path
{
    return
        _allowedTypes ?
        [_allowedTypes containsObject: [path pathExtension]] : YES;
}

- (void) removeAllObjects
{
    [_files removeAllObjects];
    _pos = 0;
}

- (NSArray *) allowedTypes
{
    return [[_allowedTypes retain] autorelease];
}

- (void) setAllowedTypes: (NSArray *) allowedTypes
{
    [_allowedTypes release];
    _allowedTypes = [allowedTypes retain];
}

- (unsigned) count
{
  return [_files count];
}

- (NSString *) currentPath
{
    NSString * res = ([_files count] > 0) ? [_files objectAtIndex: _pos] : nil;
    return res;
}

- (NSString *) currentDir
{
    return [[self currentPath] stringByDeletingLastPathComponent];
}

- (NSString *) currentName
{
    return [[self currentPath] lastPathComponent];
}

// - (NSString *) nextPath
// {
//     NSAssert(_files, @"nextPath: '_files' is not initialized");
//     NSAssert2(_pos < [_files count],
//               @"pos=%d >= count=%d", _pos, [_files count]);

//     ++_pos;
//     if (_pos >= [_files count])
//         _pos = 0;

//     return [self currentPath];
// }

// - (NSString *) prevPath
// {
//     NSAssert(_files, @"prevPath: '_files' is not initialized");
//     NSAssert2(_pos < [_files count],
//               @"pos=%d >= count=%d", _pos, [_files count]);

//     if (_pos == 0)
//     {
//         if ([_files count] > 0)
//             _pos = [_files count] - 1;
//     }
//     else
//     {
//         --_pos;
//     }

//     return [self currentPath];
// }

- (NSString *) nextPath
{
    NSAssert(_files, @"nextPath: '_files' is not initialized");
    NSAssert2([_files count] == 0 || _pos < [_files count],
              @"pos=%d >= count=%d", _pos, [_files count]);

    NSFileManager * fmgr = [NSFileManager defaultManager];


    // next in ring
    ++_pos;
    if (_pos >= [_files count])
        _pos = 0;

    while ([_files count] && [fmgr fileExistsAtPath:[self currentPath]] == NO)
    {
        [_files removeObjectAtIndex: _pos];
        if ( _pos >= [_files count] )
            _pos = 0;
    }
    
    //NSLog(@"nextPath: returning %@\n", [self currentPath]);
    return [self currentPath];
}

- (NSString *) prevPath
{
    NSAssert(_files, @"prevPath: '_files' is not initialized");
    NSAssert2([_files count] == 0 || _pos < [_files count],
              @"pos=%d >= count=%d", _pos, [_files count]);

    NSFileManager * fmgr = [NSFileManager defaultManager];

    if (_pos == 0)
        _pos = [_files count] ? [_files count] - 1 : 0;
    else
        --_pos;

    while ([_files count] && [fmgr fileExistsAtPath:[self currentPath]] == NO)
    {
        [_files removeObjectAtIndex: _pos];

        if (_pos == 0)
            _pos = [_files count] ? [_files count] - 1 : 0;
        else
            --_pos;
    }

    //NSLog(@"prevPath: returning %@\n", [self currentPath]);
    return [self currentPath];
}

@end

NSString * PhMakePath(NSString * dir, NSString * path)
{
    if ( [path isAbsolutePath] == NO )
        path = [dir stringByAppendingPathComponent: path];

    return [path stringByStandardizingPath];
}

NSString * PhMakeAbsPath(NSString * path)
{
    if ( [path isAbsolutePath] == NO )
    {
        NSFileManager * fmgr = [NSFileManager defaultManager];
        path =
            [[fmgr currentDirectoryPath] stringByAppendingPathComponent: path];
    }

    return [path stringByStandardizingPath];
}

void PhPrintArray(NSArray * arr)
{
    NSEnumerator * typesEnum = [arr objectEnumerator];
    id type;
    while ( (type = [typesEnum nextObject]) != nil )
        NSLog(@"type=%s\n", [type cString]);
}
