#include "ImageWindow.h"
#include <Foundation/NSData.h>
#include <AppKit/NSApplication.h>
#include <AppKit/NSImageView.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSScreen.h>

#include "PictureView.h"
#include "PHImage.h"
#include "Fragment.h"
#include "AppController.h"
#include "SelectionCtrl.h"
#include "LosslessJPEG.h"


#define APP_NAME "PhotoClip"
#define WIDTH_SCREEN_FRACTION (4./5.)
#define HEIGHT_SCREEN_FRACTION (4./5.)

@implementation ImageWindow

- (id) init
{
    // Determine good window size.
    NSSize screenSize = [[NSScreen mainScreen] frame].size;
    NSRect rect = NSMakeRect(0,0,
                             screenSize.width * WIDTH_SCREEN_FRACTION,
                             screenSize.height * HEIGHT_SCREEN_FRACTION);
    return [self initWithContentRect: rect];
}

- (void) dealloc
{
    [_sourcePath release];
    [super dealloc];
}

- (id) initWithContentRect: (NSRect)rect
{
    unsigned int styleMask = NSTitledWindowMask
        | NSClosableWindowMask
        | NSMiniaturizableWindowMask
        | NSResizableWindowMask;

    [self initWithContentRect: rect
          styleMask: styleMask
          // backing: NSBackingStoreBuffered
          backing: NSBackingStoreRetained
          // backing: NSBackingStoreNonretained
          defer: YES];
    [self setTitle: @APP_NAME];

    NSView * view = AUTORELEASE ([PictureView new]);
    [view initWithFrame: rect];
    [self setContentView: view];

    // Make PictureView the first responder
    if ([view acceptsFirstResponder])
        [self setInitialFirstResponder: view];

    [self setDelegate: view];

    _sourcePath = nil;
    
    return self;
}

- (void) loadFile: (NSString *)fileName
{
    NSImage * newImage = [NSImage alloc];
    newImage = [newImage initWithContentsOfFile: fileName];
    if ( newImage == nil )
    {
        NSLog(@"can't load file '%@'\n", fileName);
    }

    [[self contentView] setImage: newImage];
    [newImage release];


    // Update the title
    
    NSString * title = @APP_NAME;
    NSString * fileNameShrt = [fileName lastPathComponent];

    if ( fileNameShrt != nil )
    {
        title = [title stringByAppendingFormat: @": %@", fileNameShrt];
        if ( newImage == nil )
            title = [title stringByAppendingString: @" (failed)"];
        else
            title = [title stringByAppendingFormat: @" (%dX%d)",
                           (int)[[self contentView] imageSize].width,
                           (int)[[self contentView] imageSize].height];
    }

    [self setTitle: title];

    // Save the source file (used for loseless JPEG savings)
    // Beware! the file could be deleted by the time we're ready to save it.
    ASSIGN( _sourcePath, fileName );
    
    // Print information about image

//     NSEnumerator * reps = [[newImage representations] objectEnumerator];
//     id rep;
//     while (rep = [reps nextObject])
//     {
//         fprintf(stderr, "rep=%@\n", [[rep class] description]);
//     }
}

- (NSBitmapImageFileType) sourceFileType
{
    // Source extension
    NSString * ext = [_sourcePath pathExtension];

    if ( NSOrderedSame == [ext caseInsensitiveCompare: @"tiff"] )
        return NSTIFFFileType;

    if ( NSOrderedSame == [ext caseInsensitiveCompare: @"jpg"] )
        return NSJPEGFileType;

    if ( NSOrderedSame == [ext caseInsensitiveCompare: @"png"] )
        return NSPNGFileType;

    if ( NSOrderedSame == [ext caseInsensitiveCompare: @"gif"] )
        return NSGIFFileType;

    if ( NSOrderedSame == [ext caseInsensitiveCompare: @"bmp"] )
        return NSBMPFileType;

    return -1;
}

- (BOOL) saveToFile: (NSString *) path
{
    // Source image
    NSImage * srcImg = [[self contentView] image];
    if ( srcImg == NULL )
      return NO;

    // Source extension
    NSString * srcExt = [_sourcePath pathExtension];

    Fragment * fragm = [[self contentView] fragmentToSave];

    // Destination span
    int requiredSpan = [[[NSApp delegate] selectionCtrl] span];

    
    // Do lossless JPEG transformation if we were asked to
    // and are writing to JPEG file.
    
    NSString * ext = [path pathExtension];

    BOOL attemptLLJPEG =
        ( NSOrderedSame == [ext caseInsensitiveCompare: @"jpg"] )
        && [self tryLLJPEG]
        && [fragm scaleFor: requiredSpan] == 1.0;
    if ( attemptLLJPEG )
    {
        // Check that we really can do that.
        // This method can return NO if we could not adjust coordinates
        // due to restrictions on size and aspect ratio

        NSString * errormsg;
        BOOL canDoLLJPEG = [LosslessJPEG canCropFragment: fragm
                                         fromImage: srcImg
                                         error: &errormsg];
        if ( canDoLLJPEG )
        {
            return [LosslessJPEG xformTo: path from: _sourcePath
                                 image: srcImg fragment: fragm];
        }
        else
        {
            NSString * message =
                [NSString stringWithFormat: @"Cannot perform lossless JPEG"
                          " transformation:\n%@.\n"
                          "Try to remove constrains for image size and"
                          " aspect ratio\n"
                          "and make selection again.", errormsg];
            NSRunAlertPanel( @APP_NAME, message, 0, 0, 0 );
            return NO;
        }
    }

    // This would be the common logic to save image in any output format
    // when [[NSBitmapImageRep -representationUsingType:properties:]
    // is inplemented in GNUstep.
    
    // Create the destination image

    // [-imageMadeFromRect] directly manipulates bits in bitmap rep
    // This is a workaround until I find the way to ask the GNUstep to do
    // this or GNUstep fills up the missing functionality.
    
    NSImage * destImg = [srcImg imageMadeFromRect: [fragm scene]
                                scaledToSpan: requiredSpan
                                rotatedBy: [fragm rotation]];


    /*
    // This is the normal way  - it does not work for art backend (?)
    NSImage * destImg = [srcImg imageFromRect: [fragm scene]
                         scaledToSpan: requiredSpan
                         rotatedBy: [fragm rotation]];

    NSRect rect = NSMakeRect(0,0, [destImg size].width, [destImg size].height);

    [destImg lockFocus];

    NSBitmapImageRep * rep =
        [[NSBitmapImageRep alloc] initWithFocusedViewRect: rect];

    
    [destImg unlockFocus];
    // then NSData * data = [rep TIFFRepresentation];
    
    */


    // Save according to file type.
    // While [NSBitmapImageRep -representationUsingType:properties:]
    // is not implemented do only TIFFs.
    
    if ( NSOrderedSame == [ext caseInsensitiveCompare: @"tiff"] )
    {
        NSData * data = [destImg TIFFRepresentation];
        return [data writeToFile: path atomically: NO];
    }
    else if ( NSOrderedSame == [ext caseInsensitiveCompare: @"jpg"] )
    {
        NSString * message;

        if ( NSOrderedSame != [srcExt caseInsensitiveCompare: ext] )
        {
            message =
                @"Conversion to JPEG from other formats is not implemented.";
        }
        else
        {
            message =
                @"Saving to JPEG format is not implemented for most cases.\n"
                "Try lossless JPEG (Ctrl-L) if you do not need scaling"
                " or use TIFF format.";
        }

        NSRunAlertPanel( @APP_NAME, message, 0, 0, 0 );
        return NO;
    }
    else
    {
        NSString * message =
            [NSString stringWithFormat:
                          @"Unknown file format for extension %@", ext ];
        NSRunAlertPanel( @APP_NAME, message, 0, 0, 0 );
        return NO;
    }
}

- (BOOL) canEnableLLJPEG
{
  // TRUE if it makes sense to talk about it
  return [LosslessJPEG isAvailable] && [self sourceFileType] == NSJPEGFileType;
}

- (BOOL) tryLLJPEG
{
    return
        [self canEnableLLJPEG] &&
        [[[NSApp delegate] selectionCtrl] tryLLJPEG];
}


- (void) setRotationTo: (float) degrees
{
    [[self contentView] setRotationTo: degrees];
}

@end
