#include "LosslessJPEG.h"

#include <Foundation/NSString.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSGeometry.h>
#include <Foundation/NSFileManager.h>
#include <AppKit/NSImage.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>

#include "Fragment.h"
#include "PHAffineTransform.h"
#include "PHGeometry.h"
#include "PHImage.h"

int renameFile( const char * oldpath, const char * newpath );

@interface LosslessJPEG (PrivateMethods)

+ (BOOL) forkPipeArgs1: (NSArray *) args1
                 args2: (NSArray *) args2
                 outFD: (int) newStdout
                 errFD: (int) newStderr;

+ (BOOL) forkProcess: (NSArray *) args
               outFD: (int) newStdout
               errFD: (int) newStderr;

+ (void) redirectStdoutTo: (int) newStdout stderrTo: (int) newStderr;

+ (char **) argvFromArguments: (NSArray *) args;

+ (NSString *) commandFromArguments: (NSArray *) args;


@end

typedef enum
{
    LLJPEG_Unknown = -1,
    LLJPEG_NotAvailable = 0,
    LLJPEG_Ok = 1
}
LosslessJPEGStatus;

static LosslessJPEGStatus s_lljpegStatus = LLJPEG_Unknown;


@implementation LosslessJPEG

+ (BOOL) isAvailable
{
#if defined (JPEGTRAN)

    // We already determinted the status, do not check again
    if (s_lljpegStatus != LLJPEG_Unknown)
        return ( s_lljpegStatus == LLJPEG_Ok );

    // Make pessimistic assumption
    s_lljpegStatus = LLJPEG_NotAvailable;
    
    NSFileManager * fmgr = [NSFileManager defaultManager];
    if ( ! [fmgr fileExistsAtPath: @JPEGTRAN] )
        return ( s_lljpegStatus == LLJPEG_Ok );

    if ( ! [fmgr isExecutableFileAtPath: @JPEGTRAN] )
        return ( s_lljpegStatus == LLJPEG_Ok );

    // Find sample jpeg file

    NSBundle * bun = [NSBundle mainBundle];
    NSString * sampleJpeg = [bun pathForResource: @"sample" ofType: @"jpg" ];

    if ( ! [[NSFileManager defaultManager] fileExistsAtPath: sampleJpeg] )
        return ( s_lljpegStatus == LLJPEG_Ok );
    
    // Run test jpegtran with sample jpeg file
    //   jpegtran -crop 8x8+0+0 -rotate 90 sample.jpg

    NSMutableArray * args = [NSMutableArray new];
    [args addObject: @JPEGTRAN];
    [args addObject: @"jpegtran"];
    [args addObject: @"-crop"];
    [args addObject: @"8x8+0+0"];
    [args addObject: @"-rotate"];
    [args addObject: @"90"];
    [args addObject: sampleJpeg];


    int fd = open("/dev/null", O_WRONLY);
    if ( fd == -1 )
    {
        perror( "open( /dev/null ) failed" );
        return NO; // try again next time we check
    }
        
    BOOL result = [LosslessJPEG forkProcess: args outFD: fd errFD: fd];

    [args release];

    if ( close(fd) == -1 )
        perror( "close() failed" );

    s_lljpegStatus = result ? LLJPEG_Ok : LLJPEG_NotAvailable;
    return (s_lljpegStatus == LLJPEG_Ok);
    
#else // ! defined (JPEGTRAN)

    return NO;

#endif // ! defined (JPEGTRAN)
}


+ (BOOL) canCropFragment: (Fragment *) fragm fromImage: (NSImage *) image
                   error: (NSString **) error
{
    // This method used a simplified critetia that are sufficient
    // but not nesessary: the origin should lie on block boundary
    // and in the case of rotation the fragment boundaries should
    // lie on block boundaries, too.

    // However, this restriction has already been applied in
    // [Selection -adjustToJPEGGrid] - so we can rotate the
    // selection without futher adjustment - and it makes no harm
    // to apply in again.

    NSSize blockSize = [image jpegBlockSize];
    NSSize imageSize = [image size];
    NSRect rc = [fragm scene];

    int x = (int) rint( NSMinX(rc) );
    int y = (int) rint( imageSize.height - NSMaxY(rc) );
    int dx = (int) rint( blockSize.width );
    int dy = (int) rint( blockSize.height );

    // Origin must be on block boundary

    if ( (x % dx) || (y % dy) )
    {
        if ( error)
            *error = [NSString stringWithFormat: @"upper left point (%d,%d)"
                               " is not aligned with blocks {%dx%d}",
                               x, y, dx, dy];
        return NO;
    }

    if ( [fragm rotation] != 0 )
    {
        // scene must be alighed with blocks
        int w = (int) rint( NSWidth(rc) );
        int h = (int) rint( NSHeight(rc) );

        if ( (w % dx) || (h % dy) )
        {
            if (error)
                *error = [NSString stringWithFormat: @"fragment {%dx%d}"
                                   " cannot be divided into blocks {%dx%d}",
                                   w, h, dx, dy];
            return NO;
        }
    }

    if ( error)
        *error = NULL;
    return YES;
}

+ (BOOL) xformTo: (NSString *) dest from: (NSString *) src
           image: (NSImage *) img fragment: (Fragment *) fragm
{
    static char errorbuf[256];
    
    // Current implemenation delegate all the job to jpegtran program
    // and invokes the jpegtran process if this program exists.

    if ( [LosslessJPEG isAvailable] == NO )
        return NO;

    // Take care not to overwrite the source file with empty destination.
    
    BOOL replace = (NSOrderedSame == [dest compare: src]);

    NSString * actualDest = 0;
    int fd = 0;
    if ( replace )
    {
        // Create and open temporary file

        char buffer[32];
        strcpy(buffer, "/tmp/PhotoClipXXXXXX");
        if ( (fd = mkstemp(buffer)) == -1 )
        {
            sprintf( errorbuf, "mkstemp(%s) failed", buffer );
            perror( errorbuf );
            return NO;
        }
        actualDest = [[NSString alloc] initWithCString: buffer];
    }
    else
    {
        // Create and open destination file

        if ((fd = open([dest cString], O_WRONLY|O_CREAT|O_TRUNC, 0666)) == -1)
        {
            sprintf( errorbuf, "open(%s) failed", [dest cString] );
            perror( errorbuf );
            return NO;
        }
        actualDest = [dest retain];
    }

    
    // Create command line.

    // Although the general jpegtran syntax is
    //   jpegtran [-crop WxH+X+Y] [-rotate <90|180|270>] [<input-file>],
    // we create one of three variants:
    // 1. No rotation:
    //      jpegtran -crop WxH+X+Y <input-file>
    // 2. No cropping:
    //      jpegtran -rotate <90|180|270> <input-file>
    // 3. For both cropping and rotation we create a pipe,
    //    first crop, then rotate, the order is important:
    //      jpegtran -crop WxH+X+Y <input-file> | jpegtran -rotate <90|180|270>

    NSMutableArray * args = [[NSMutableArray alloc] init];
    NSMutableArray * cropArgs = 0;
    NSMutableArray * rotateArgs = 0;

    [args addObject: @JPEGTRAN]; // executable path

    NSString * program =
        [[NSString stringWithString: @JPEGTRAN] lastPathComponent];
    [args addObject: program];   // argv[0]

    NSRect rc = [fragm scene];
    NSRect rcImage = [img rect];
    int degrees = [fragm rotation];
    BOOL doCrop = ( NSEqualRects( rcImage, rc ) == NO );
    BOOL doRotate = ( degrees != 0 );

    if ( doCrop  )
    {
        cropArgs = [[NSMutableArray alloc] init];
        [cropArgs addObject: @"-crop"];
        [cropArgs addObject: [NSString stringWithFormat: @"%gx%g%+g%+g",
                                       rint( NSWidth(rc) ),
                                       rint( NSHeight(rc) ),
                                       rint( NSMinX(rc) ),
                                       rint(NSHeight(rcImage) - NSMaxY(rc))]];
    }
    

    if ( doRotate )
    {
        rotateArgs = [[NSMutableArray alloc] init];
        [rotateArgs addObject: @"-rotate"];
        [rotateArgs addObject: [NSString stringWithFormat: @"%d",
                                         360 - degrees]];
    }

    BOOL result = NO;

    if ( doCrop == NO || doRotate == NO )
    {
        if ( doCrop )
            [args addObjectsFromArray: cropArgs];
        else if ( doRotate )
            [args addObjectsFromArray: rotateArgs];

        [args addObject: src];

        result = [LosslessJPEG forkProcess: args outFD: fd errFD: 2 ];
    }
    
    if ( doCrop == YES && doRotate == YES )
    {
        [args addObjectsFromArray: cropArgs];
        [args addObject: src];

        NSMutableArray * args2 = [[NSMutableArray alloc] init];
        [args2 addObject: @JPEGTRAN]; // executable path
        [args2 addObject: program];   // argv[0]
        [args2 addObjectsFromArray: rotateArgs];

        result = [LosslessJPEG forkPipeArgs1: args args2: args2
                               outFD: fd errFD: 2 ];

        [args2 release];
    }

    [args release];

    if ( close( fd ) == -1 )
    {
        sprintf( errorbuf, "close(%d) failed", fd );
        perror( errorbuf );
        result = NO;
    }

    if ( result == YES && replace == YES )
    {
        // Rename the file but first check that actualDest exists

        result = NO;
        struct stat bufstat;
        if ( stat( [actualDest cString], &bufstat ) == 0 )
            if ( bufstat.st_size > 0 )
                if ( renameFile( [actualDest cString], [dest cString] ) == 0 )
                    result = YES;
                    

    }

    [actualDest release];
    return result;
}

+ (BOOL) forkProcess: (NSArray *) args
               outFD: (int) newStdout
               errFD: (int) newStderr
{
    static char buffer[256];
    BOOL result = NO;

    //NSLog( @"Executing %@", [LosslessJPEG commandFromArguments: args] );

    // Fork the child process
    
    pid_t pid = fork();
    if ( pid == -1 )
    {
        perror( "fork() failed" );
        return result;
    }
    else if ( pid == 0 )
    {
        // Child

        char ** argv = [LosslessJPEG argvFromArguments: args ];
        if ( ! argv )
        {
            fprintf( stderr, "argv allocation failed\n" );
            exit(1);
        }

        [LosslessJPEG redirectStdoutTo: newStdout stderrTo: newStderr];

        // Execute jpegtran
        // argv[0] contains the executable path,
        // argv[1] will be seen by the new executable as argv[0].
        execve( argv[0], argv+1, 0 );

        // Should not get here
        sprintf( buffer, "execve(%s) failed", argv[0] );
        perror( buffer );
        exit( 1 );
    }

    // Parent continues.

    // Wait for the child to terminate
    int status = 0;
    pid_t ret;
    while ( (ret = waitpid( pid, &status, 0 )) == -1 && errno == EINTR )
        ;
    if ( ret == -1 )
    {
        perror( "waitpid() failed" );
        status = -1;
    }

    if ( WIFEXITED(status) && WEXITSTATUS(status) == 0 )
        result = YES; // successful completion

    return result;
}

+ (BOOL) forkPipeArgs1: (NSArray *) args1
                 args2: (NSArray *) args2
                 outFD: (int) newStdout
                 errFD: (int) newStderr
{
    static char buffer[256];
    BOOL result = NO;

    //NSLog( @"Executing %@ | %@",
    //       [LosslessJPEG commandFromArguments: args1],
    //       [LosslessJPEG commandFromArguments: args2] );
    
    int pipefd[2];
    if ( pipe( pipefd ) == -1 )
    {
        perror( "pipe() failed" );
        return result;
    }

    // Fork process #2 (reader)

    pid_t pid2 = fork();
    if ( pid2 == -1 )
    {
        perror( "fork failed" );
        return result;
    }
    else if ( pid2 == 0 )
    {
        // Child -> reader process
        
        if ( dup2( pipefd[0], 0 ) == -1 )
        {
            sprintf( buffer, "dup2(%d, %d) failed", pipefd[0], 0 );
            perror( buffer );
            exit( 1 );
        }

        if ( close( pipefd[0] ) == -1 )
        {
            sprintf( buffer, "close(%d) failed", pipefd[0] );
            perror( buffer );
            exit( 1 );
        }

        if ( close( pipefd[1] ) == -1 )
        {
            sprintf( buffer, "close(%d) failed", pipefd[1] );
            perror( buffer );
            exit( 1 );
        }

        // Execute reader program
        
        char ** argv = [LosslessJPEG argvFromArguments: args2];
        if ( ! argv )
        {
            fprintf( stderr, "argv allocation failed\n" );
            exit(1);
        }

        [LosslessJPEG redirectStdoutTo: newStdout stderrTo: newStderr];
        
        // argv[0] contains the executable path,
        // argv[1] will be seen by the new executable as argv[0].
        execve( argv[0], argv+1, 0 );

        // Should not get here
        sprintf( buffer, "execve(%s) failed", argv[0] );
        perror( buffer );
        exit( 1 );
    }

    // Parent continues

    // Fork process #1 (writer)

    pid_t pid1 = fork();
    if ( pid1 == -1 )
    {
        perror( "fork failed" );
        return result;
    }
    else if ( pid1 == 0 )
    {
        // Child -> writer process
        
        if ( dup2( pipefd[1], 1 ) == -1 )
        {
            sprintf( buffer, "dup2(%d, %d) failed", pipefd[1], 1 );
            perror( buffer );
            exit( 1 );
        }

        if ( close( pipefd[0] ) == -1 )
        {
            sprintf( buffer, "close(%d) failed", pipefd[0] );
            perror( buffer );
            exit( 1 );
        }

        if ( close( pipefd[1] ) == -1 )
        {
            sprintf( buffer, "close(%d) failed", pipefd[1] );
            perror( buffer );
            exit( 1 );
        }

        // Execute writer program
        
        char ** argv = [LosslessJPEG argvFromArguments: args1];
        
        // argv[0] contains the executable path,
        // argv[1] will be seen by the new executable as argv[0].
        execve( argv[0], argv+1, 0 );

        // Should not get here
        sprintf( buffer, "execve(%s) failed", argv[0] );
        perror( buffer );
        exit( 1 );
    }

    // Parent continues

    // Close pipe descriptors
    close( pipefd[0] );
    close( pipefd[1] );
    
    // Wait for both children to terminate
    int status1 = 0;
    int status2 = 0;
    pid_t ret;
    while ( (ret = waitpid( pid1, &status1, 0 )) == -1 && errno == EINTR )
        ;
    if ( ret == -1 )
    {
        perror( "waitpid() for writer process failed" );
        status1 = -1;
    }
    
    while ( (ret = waitpid( pid2, &status2, 0 )) == -1 && errno == EINTR )
        ;
    if ( ret == -1 )
    {
        perror( "waitpid() for reader process failed" );
        status2 = -1;
    }

    if ( WIFEXITED(status1) && WEXITSTATUS(status1) == 0 )
        if ( WIFEXITED(status2) && WEXITSTATUS(status2) == 0 )
            result = YES; // successful completion
    
    return result;
}
    

+ (void) redirectStdoutTo: (int) newStdout stderrTo: (int) newStderr
{
    static char buffer[256];
    
    if ( newStdout != 1 )
    {
        // Point stdout to newStdout
        if ( dup2( newStdout, 1 ) == -1 )
        {
            sprintf( buffer, "dup2(%d, %d) failed", newStdout, 1 );
            perror( buffer );
            exit( 1 );
        }

        if ( close( newStdout ) == -1 )
        {
            sprintf( buffer, "close(%d) failed", newStdout );
            perror( buffer );
            exit( 1 );
        }
    }
      
    if ( newStderr != 2 )
    {
        if ( newStderr!= newStdout )
        {
            // Point stderr to newStderr
            if ( dup2( newStderr, 2 ) == -1 )
            {
                sprintf( buffer, "dup2(%d, %d) failed", newStderr, 2 );
                perror( buffer );
                exit( 1 );
            }

            if ( close( newStderr ) == -1 )
            {
                sprintf( buffer, "close(%d) failed", newStderr );
                perror( buffer );
                exit( 1 );
            }
        }
        else
        {
            // stderrPath is the same as stdoutPath. Map stderr to stdout.
            if ( dup2( 1, 2 ) == -1 )
            {
                perror("dup2(1, 2) failed");
                exit( 1 );
            }
        }
    }
}

+ (char **) argvFromArguments: (NSArray *) args
{
    char ** argv = malloc( ([args count] + 1) * sizeof(char *) );
    int i;
    for ( i = 0; i < [args count]; ++i )
        argv[i] = (char *) [[args objectAtIndex: i] cString];

    argv[i] = 0;

    return argv;
}

+ (NSString *) commandFromArguments: (NSArray *) args
{
    NSMutableString * command = [NSMutableString string];
    
    int i;
    for ( i = 1; i < [args count]; ++i )
    {
        if ( i != 1 )
            [command appendString: @" "];
        [command appendString: [args objectAtIndex: i]];
    }

    return command;
}

@end
