#include "MouseTracker.h"
#include <AppKit/NSCursor.h>
#include "PHGeometry.h"
#include "AppController.h"
#include "SelectionCtrl.h"
#include "UnconstrainedRectCalculator.h"
#include "ConstrainedRectCalculator.h"

#define SIGN(x) ((x) > 0 ? 1 : -1)

#define INTEGRAL_SIZE

typedef enum { PHQuadOne, PHQuadTwo, PHQuadThree, PHQuadFour } PHQuadrant;

static const float TOLERANCE = 7; // tolerance = 7 pixels in sketch


@interface MouseTracker (ProtectedMethods)

- (float) requestedShape;

@end // MouseTracker (ProtectedMethods)


@implementation MouseTracker

- (id) initWithFrame: (NSRect) frame selRect: (NSRect) selRect
               start: (NSPoint) pt
{
    [self initWithFrame: frame];
    
    // ignore selRect

    [self mouseDown: pt];

    return self;
}

- (id) initWithFrame: (NSRect) frame;
{
    [super init];
    _frame = frame;
    _shape = [self requestedShape];

    return self;
}

- (NSRect) mouseDown: (NSPoint) pt
{
    return NSZeroRect;
}

- (NSRect) mouseDragged: (NSPoint) pt
{
    return NSZeroRect;
}

- (NSRect) selectionRect
{
    return NSZeroRect;
}

- (NSCursor *) cursor
{
    return [[self class] cursorType];
}

- (NSRect) frame
{
    return _frame;
}

- (float) shape
{
    return _shape;
}

- (float) requestedShape
{
    const float MIN_SHAPE = 0.01; // minimal dy/dx ratio
    
    AppController * app_ctrl = [NSApp delegate];
    float shape = [[app_ctrl selectionCtrl] shape];

    if (shape < MIN_SHAPE)
        shape = 0;

    return shape;
}

@end //   MouseTracker


// MouseTrackerNewSel ---------------------------------------------------------

@interface MouseTrackerNewSel (ProtectedMethods)

- (NSPoint) correctEndPointFromStart: (NSPoint) start end: (NSPoint) end;

@end // MouseTrackerNewSel (ProtectedMethods)


@implementation MouseTrackerNewSel

- (void) dealloc
{
    RELEASE(_calc);
    [super dealloc];
}

- (NSRect) mouseDown: (NSPoint) pt
{
    _start = PHClosestPointInRect(pt, _frame);

#ifdef INTEGRAL_SIZE
    // truncate fractional part
    _start = PHIntegralPoint(_start);
#endif // INTEGRAL_SIZE
    
    _end = _start;

    // Create RectCalculator

    Class cls;
    cls = (_shape <= 0) ?
        [UnconstrainedRectCalculator class] :
        [ConstrainedRectCalculator class];

    ASSIGN(_calc, [[cls alloc] initWithTracker: self mouse: pt]);
     
    return PHRectFromPoints(_start, _end);
}

- (NSRect) mouseDragged: (NSPoint) pt
{
    if (NSIntersectsRect(_frame, PHRectFromPoints(_start, pt)) == NO)
        return NSZeroRect; // invalid selection
    
    pt = PHClosestPointInRect(pt, _frame);
    _end = [self correctEndPointFromStart: _start end: pt];

    if (PHDistance(_start, _end) <= TOLERANCE)
        _end = _start; // invalidate selection

    return PHRectFromPoints(_start, _end);
}

- (NSRect) selectionRect
{
    return PHRectFromPoints(_start, _end);
}

- (NSPoint) correctEndPointFromStart: (NSPoint) start end: (NSPoint) end
{
    NSPoint result = [_calc correctEndPointFromStart: start end: end];

#ifdef INTEGRAL_SIZE
    result = PHIntegralPoint(result);
#endif // INTEGRAL_SIZE

    return result;
}

+ (NSCursor *) cursorType
{
    return [NSCursor greenArrowCursor];
}

@end // MouseTrackerNewSel

// MouseTrackerMoveCorner -----------------------------------------------------

@implementation MouseTrackerMoveCorner

- (id) initWithFrame: (NSRect) frame selRect: (NSRect) selRect
               start: (NSPoint) pt
{
    [super initWithFrame: frame];
    _start = selRect.origin;
    _end = PHMovePoint(_start, selRect.size.width, selRect.size.height);
    [self mouseDown: pt];
    return self;
}

- (NSRect) mouseDown: (NSPoint) pt
{
    // Determine the end and start points.
    // End point is the corner point closest to pt,
    // start - the opposite corner.

    NSRect selRect = PHRectFromPoints(_start, _end);
//     float mid_x = NSMidX(selRect);
//     float mid_y = NSMidY(selRect);
//     float min_x = NSMinX(selRect);
//     float min_y = NSMinY(selRect);

    BOOL x_pos = (pt.x > NSMidX(selRect));
    BOOL y_pos = (pt.y > NSMidY(selRect));
    PHQuadrant quad = (x_pos) ?
        (y_pos ? PHQuadOne : PHQuadFour) : (y_pos ? PHQuadTwo : PHQuadThree);

    switch (quad)
    {
        case PHQuadOne:
            _end.x = NSMaxX(selRect);
            _end.y = NSMaxY(selRect);
            _start.x = NSMinX(selRect);
            _start.y = NSMinY(selRect);
            break;
        case PHQuadTwo:
            _end.x = NSMinX(selRect);
            _end.y = NSMaxY(selRect);
            _start.x = NSMaxX(selRect);
            _start.y = NSMinY(selRect);
            break;
        case PHQuadThree:
            _end.x = NSMinX(selRect);
            _end.y = NSMinY(selRect);
            _start.x = NSMaxX(selRect);
            _start.y = NSMaxY(selRect);
            break;
        case PHQuadFour:
            _end.x = NSMaxX(selRect);
            _end.y = NSMinY(selRect);
            _start.x = NSMinX(selRect);
            _start.y = NSMaxY(selRect);
            break;
        default:
            NSAssert1(0, @"invalid quadrant=%d", quad);
    }

    // Create RectCalculator
    Class cls = (_shape <= 0) ?
        [UnconstrainedRectCalculator class] :
        [ConstrainedRectCalculator class];

    ASSIGN(_calc, [[cls alloc] initWithTracker: self mouse: pt
                               start: _start end: _end]);

    _pt_to_end = [_calc mouseToEndOffset];

    return selRect;
}

- (NSRect) mouseDragged: (NSPoint) pt
{
    pt = PHMovePointByVect(pt, _pt_to_end);
    _end = [self correctEndPointFromStart: _start end: pt];

    if (PHDistance(_start, _end) <= TOLERANCE)
        _end = _start; // invalidate selection

    return PHRectFromPoints(_start, _end);
}

+ (NSCursor *) cursorType
{
    return [NSCursor greenArrowCursor];
}

@end // MouseTrackerMoveCorner


// MouseTrackerMoveAll --------------------------------------------------------


@implementation MouseTrackerMoveAll

- (id) initWithFrame: (NSRect) frame selRect: (NSRect) selRect
               start: (NSPoint) pt
{
    [super initWithFrame: frame];

    _selRect = selRect;

    [self mouseDown: pt];
    return self;
}

- (NSRect) mouseDown: (NSPoint) pt
{
    _start = pt;
    return _selRect;
}

- (NSRect) mouseDragged: (NSPoint) pt
{
    _selRect = NSOffsetRect(_selRect, pt.x - _start.x, pt.y - _start.y);
    _selRect = PHPushRectInsideBoundaries(_selRect, _frame);

    _start = pt;

    return _selRect;
}

- (NSRect) selectionRect
{
    return _selRect;
}

+ (NSCursor *) cursorType
{
    // TODO: make appropriate cursor, probably "+"-arrows
    return [NSCursor arrowCursor];
}

@end // MouseTrackerMoveAll
