#include "PHImage.h"
#include <Foundation/NSException.h>
#include "PHGeometry.h"
#include "PHAffineTransform.h"
#include "mem.h"
#include "bmpcomponent.h"

#include <math.h>


@implementation NSImage (PhotoClip)

- (NSRect) rect
{
    NSRect rc;
    rc.origin = NSZeroPoint;
    rc.size = [self size];
    return rc;
}

- (NSSize) jpegBlockSize
{
    // TODO: Need to extract this information from the JPEG image header
    return NSMakeSize( 16, 16 );
}

- (NSBitmapImageRep *) bitmapRepresentation
{
    // Find NSBitmapImageRep among existing representation
    Class bitmapClass = [NSBitmapImageRep class];

    NSEnumerator * reps = [[self representations] objectEnumerator];
    NSImageRep * rep;
    while ( (rep = [reps nextObject]) != nil )
        if ( [rep isKindOfClass: bitmapClass] )
            return (NSBitmapImageRep *)rep;

    // Not found, create one (?)
    return [NSBitmapImageRep imageRepWithData: [self TIFFRepresentation]];
}

- (NSImage *) imageFromRect: (NSRect) rect
{
    NSImage * canvas = [[NSImage alloc] initWithSize: rect.size];
    [canvas lockFocus];

    [self compositeToPoint: NSZeroPoint
          fromRect: rect
          operation: NSCompositeCopy];

    [canvas unlockFocus];
    return [canvas autorelease];
}

- (NSImage *) imageTransformedBy: (NSAffineTransform *) transform
{
    NSSize new_size = [transform transformSize: [self size]];
    NSImage * canvas = [[NSImage alloc] initWithSize: new_size];
    [canvas lockFocus];

    [transform concat];

    // Get NSImageRep of image
    NSImageRep * rep = [self bestRepresentationForDevice: nil];

    [rep drawAtPoint: NSZeroPoint];
    
    [canvas unlockFocus];
    return [canvas autorelease];
}


// This method (transformation after clip) does not work

- (NSImage *) imageFromRect: (NSRect) rect
              transformedBy: (NSAffineTransform *) transform
{
    NSRect original_rect;
    original_rect.origin = NSZeroPoint;
    original_rect.size = [self size];

    if ( NSEqualRects(original_rect, NSIntersectionRect(original_rect, rect)) )
        return [self imageTransformedBy: transform]; // nothing to clip

    // First, clip the scene
    
    NSImage * fragment = [[NSImage alloc] initWithSize: rect.size];
    [fragment lockFocus];

    [self compositeToPoint: NSZeroPoint
          fromRect: rect
          operation: NSCompositeCopy];

    [fragment unlockFocus];

    // Then, transform.
    NSImage * result = [fragment imageTransformedBy: transform];
    RELEASE(fragment);
    
    return result;
}

- (NSImage *) imageFromRect: (NSRect) rect
                  rotatedBy: (float) degrees
{
    return [self imageFromRect: rect scaledBy: 1.0 rotatedBy: degrees];
}

- (NSImage *) imageFromRect: (NSRect) rect
                   scaledBy: (float) scale
                  rotatedBy: (float) degrees
{
    NSAffineTransform * xform = [NSAffineTransform transformToQIofRect: rect
                                                   scale: scale
                                                   rotation: degrees];
    
    NSSize canvas_size = [xform transformSize: rect.size];

    canvas_size.width = floorf(canvas_size.width);
    canvas_size.height = floorf(canvas_size.height);

    NSImage * canvas = [[NSImage alloc] initWithSize: canvas_size];
    [canvas lockFocus];

    [xform concat];
    
    // Get NSImageRep of image
    NSImageRep * rep = [self bestRepresentationForDevice: nil];

    [rep drawAtPoint: NSZeroPoint];
    
    [canvas unlockFocus];
    return AUTORELEASE(canvas);
}


// Performs image bits recalculation (instead of asking GNUstep to do that)

- (NSImage *) imageMadeFromRect: (NSRect) rect
                   scaledToSpan: (int) span
                      rotatedBy: (float) degrees
{
    // Fix source rectangle

    rect = PHNearestIntegerRect( rect );
    rect = NSIntersectionRect( rect, [self rect] );
    
    // Determine scaled size and scale

    float scale = 1.0;
    NSSize scaled_size = PHScaledSize( rect.size, span, &scale );

    NSSize dest_size = PHRotateSize( scaled_size, (int)degrees );


    NSBitmapImageRep * src_rep;
    NSBitmapImageRep * dest_rep;

    src_rep = [self bitmapRepresentation];
    
    
    NSString * colorSpaceName = [src_rep colorSpaceName];
    BOOL hasAlpha = [src_rep hasAlpha];
    BOOL isPlanar = [src_rep isPlanar];
    //BOOL isOpaque = [src_rep isOpaque];
    int bitsPerSample = [src_rep bitsPerSample];
    int samplesPerPixel = [src_rep samplesPerPixel];

    //int bitsPerPixel = [src_rep bitsPerPixel];
    //int numberOfPlanes = [src_rep numberOfPlanes] ;
    //int bytesPerPlane = [src_rep bytesPerPlane];
    //int bytesPerRow =  [src_rep bytesPerRow];

    //NSLog(@"colorSpaceName=%@ hasAlpha=%d isOpaque=%d",
    //      colorSpaceName, hasAlpha, isOpaque);
    //NSLog(@"bitsPerSample=%d bitsPerPixel=%d samplesPerPixel=%d isPlanar=%d",
    //      bitsPerSample, bitsPerPixel, samplesPerPixel, isPlanar);
    //NSLog(@"isPlanar=%d numberOfPlanes=%d bytesPerPlane=%d bytesPerRow=%d",
    //      isPlanar, numberOfPlanes, bytesPerPlane, bytesPerRow);

    // We can process further only if isPlanar == NO and hasAlpha == NO
    if (hasAlpha == YES || isPlanar == YES)
    {
        // TODO: exception
        if (hasAlpha == YES)
            NSLog(@"hasAplha == YES is not implemented");
        if (isPlanar == YES)
            NSLog(@"isPlanar == YES is not implemented");
        return nil;
    }
    

    // Allocate destination image and its representation

    NSImage * dest_image = [[NSImage alloc] initWithSize: dest_size];
    dest_rep = [[NSBitmapImageRep alloc]
                   initWithBitmapDataPlanes:NULL
                   pixelsWide: dest_size.width
                   pixelsHigh: dest_size.height
                   bitsPerSample:bitsPerSample
                   samplesPerPixel:samplesPerPixel
                   hasAlpha:hasAlpha
                   isPlanar:isPlanar
                   colorSpaceName:colorSpaceName
                   bytesPerRow: 0
                   bitsPerPixel: 0
    ];


    [self moveBitsFromSrcRep: src_rep srcRect: rect
          toDestRep: dest_rep scaleBy: scale rotateBy: degrees];
    
    [dest_image addRepresentation: dest_rep];
    [dest_rep release];
    
    return [dest_image autorelease];
}


#define DEST_INDEX( x, y, w, h, rotation, x_dest, y_dest) \
{                                   \
    switch (rotation)               \
    {                               \
        case 0:                     \
            x_dest = x;             \
            y_dest = y;             \
            break;                  \
        case 90:                    \
            x_dest = y;             \
            y_dest = w - 1 - x;     \
            break;                  \
        case 180:                   \
            x_dest = w - 1 - x;     \
            y_dest = h - 1 - y;     \
            break;                  \
        case 270:                   \
            x_dest = h - 1 - y;     \
            y_dest = x;             \
            break;                  \
    }                               \
}                                   \


- (void) moveBitsFromSrcRep: (NSBitmapImageRep *) src
                    srcRect: (NSRect) srcRect
                  toDestRep: (NSBitmapImageRep *) dest
                    scaleBy: (float) scale
                   rotateBy: (float) degrees
{

    NSAssert( scale <= 1, @"scale must be less than 1");

    // Extract scaledSize from dest
    NSSize scaledSize = NSMakeSize( [dest pixelsWide], [dest pixelsHigh] );
    int rotation = (int)degrees;

    scaledSize = PHRotateSize( scaledSize, rotation );

    // delta is the size of destination pixel measured in source pixels.
    // delta = 1/scale, since the size if source pixel is 1.
    // We'll write the code that accepts different delta.x and delta.y
    // but assign delta.x = delta.y in the very beginning

    NSPoint delta;
    delta.x = 1.0/scale;
    delta.y = 1.0/scale; // now the same, but can be different


    //NSLog(@"moveBitsFromSrcRep: srcRect=%@", NSStringFromRect( srcRect ));
    //NSLog(@"moveBitsFromSrcRep: delta=(%f, %f)", delta.x, delta.y);


    int h = [src pixelsHigh];
    int w = [src pixelsWide];
    int hh =  scaledSize.height;
    int ww =  scaledSize.width;
    //int hhh = [dest pixelsHigh];
    int www = [dest pixelsWide];

    //NSLog( @"h=%d w=%d hh=%d ww=%d\n", h, w, hh, ww);

    int x, y; // source bitmap indexes
    int i, j; // source bitmap indexes in the src rect, they start from 0
    int c;    // byte count in a pixel, ~color
    
    int y_top = NSMaxY( srcRect );
    int y_bot = NSMinY( srcRect );

    // bitmap rows go from top to bottom
    int y_start = h - y_top;
    int y_end = h - y_bot;
    int x_start = NSMinX( srcRect );
    int x_end = NSMaxX( srcRect );

    //NSLog(@"y_start=%d, y_end=%d, x_start=%d, x_end=%d\n",
    //      y_start, y_end, x_start, x_end);

    unsigned char * src_data = [src bitmapData];
    unsigned char * dest_data = [dest bitmapData];

    // Destination bitmap was created with the same color depth as source.

    int bitsPerSample = [src bitsPerSample];
    int samplesPerPixel = [src samplesPerPixel];
    int bitsPerPixel = [src bitsPerPixel];

    if ( bitsPerPixel % 8 != 0 )
    {
        NSLog(@"bitsPerPixel == %d not implemented", bitsPerPixel);
        return;
    }
            
    int bytesPerPixel = bitsPerPixel / 8;

    // NSLog( @"bitsPerSample=%d samplesPerPixel=%d, bytesPerPixel=%d\n",
    //       bitsPerSample, samplesPerPixel, bytesPerPixel);

    if ( delta.x == 1.0 && delta.y == 1.0 )
    {
        // Simpler processing for no scale condition

        int src_pixel, dest_pixel;
        int xx, yy;

        for ( j=0, y = y_start; y < y_end; ++y, ++j )
        {
            for ( i=0, x = x_start; x < x_end; ++x, ++i )
            {
                // caclulate destination indexes xx, yy
                DEST_INDEX( i, j, ww, hh, rotation, xx, yy);

                dest_pixel = yy * www + xx;
                src_pixel = y * w + x;
                for (c = 0; c < bytesPerPixel; ++c)
                    dest_data[dest_pixel * bytesPerPixel + c] =
                        src_data[src_pixel * bytesPerPixel + c];
            }
        }
        
        return;
    }

    NSAssert( bitsPerSample * samplesPerPixel <= bitsPerPixel,
              @"bitsPerSample * samplesPerPixel > bitsPerPixel");

    // Destination matrix for components (RGB, RGBA, CMYK, CMYKA).
    // The size must be on pixel greater than hh and ww due to the algorithm:
    // last source pixels might not be completely covered
    // by destinatrion pixels.

    float *** m = fmatrix3d( samplesPerPixel, hh+1, ww+1 );

    // Scale source rectangle to matrix
    
    int xx1, xx2, yy1, yy2; // dest indexes
    float a1, a2, b1, b2;   // weights

    int src_pixel;
    float value;            // color value of source pixel

    for ( c=0; c < samplesPerPixel; ++c )
    {
        // Each source pixel can fall into one, two or four destination
        // pixels. Go over source pixels and calculate the fraction that
        // falls into each destination pixel.
        for ( j = 0, y = y_start; y < y_end; ++y, ++j )
        {
            yy1 = (int)( j/delta.y );
            yy2 = (int)( (j+1)/delta.y );
            b1 = (yy2 * delta.y - j);
            if ( b1 < 0 )
                b1 = 0;
            b2 = 1 - b1;

            for ( i = 0, x = x_start; x < x_end; ++x, ++i )
            {
                xx1 = (int)( i/delta.x );
                xx2 = (int)( (i+1)/delta.x );
                a1 = (xx2 * delta.x - i);
                if ( a1 < 0 )
                    a1 = 0;
                a2 = 1 - a1;

                //fprintf(stdout, "(c, i, j)=(%d, %d, %d)\n", c, i, j);
                //fprintf(stdout, "xx1=%d, xx2=%d, yy1=%d, yy2=%d\n",
                //        xx1, xx2, yy1, yy2);
                //fprintf(stdout, "a1=%f, a2=%f, b1=%f, b2=%f\n",
                //        a1, a2, b1, b2);

                src_pixel = bytesPerPixel * (y*w + x);

                // good only if bitsPerSample == 8
                //value = src_data[ src_pixel + c ];

                // COMPONENT_VALUE supports several bps, but slower
                COMPONENT_VALUE(src_data + src_pixel, c, bitsPerSample, value);
                
                if ( a1 == 0 && b1 == 0 )
                {
                    // an attempt to speedup
                    m[c][yy1][xx1] += value;
                }
                else
                {
                    m[c][yy1][xx1] += value * b1 * a1;
                    m[c][yy1][xx2] += value * b1 * a2;
                    m[c][yy2][xx1] += value * b2 * a1;
                    m[c][yy2][xx2] += value * b2 * a2;
                }
            }   
        }
    }


    // Normalize destination matrix and fetch into destination image rep

    double square = delta.x * delta.y;

    // NSLog(@"square=%lf\n", square);

    int dest_pixel;
    for ( c = 0; c < samplesPerPixel; ++c )
    {
        for ( j = 0; j < hh; ++j )
            for ( i = 0; i < ww; ++i )
            {
                // calculate destination indexes x, y
                DEST_INDEX( i, j, ww, hh, rotation, x, y);

                dest_pixel = bytesPerPixel * (y * www + x);

                // good only if bitsPerSample == 8
                //dest_data[dest_pixel + c] = (m[c][j][i] / square);

                // SET_COMPONENT_VALUE supports several bitsPerSample but slow
                SET_COMPONENT_VALUE( dest_data + dest_pixel, c, bitsPerSample,
                                     (m[c][j][i] / square) );
            }
    }

    //NSLog(@"before free_fmatrix3d()");
    
    free_fmatrix3d( m, samplesPerPixel, hh + 1, ww + 1 );
}

@end
