#include "PictureView.h"

#include <Foundation/NSException.h>
#include <Foundation/NSGeometry.h>
#include <Foundation/NSData.h>
#include <AppKit/NSAffineTransform.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSEvent.h>
#include <AppKit/NSWindow.h>
#include <AppKit/NSApplication.h>
#include <AppKit/NSMenu.h>
#include <AppKit/NSCursor.h>
#include <math.h>

#include "PHImage.h"
#include "PHAffineTransform.h"
#include "Selection.h"
#include "SelectionCtrl.h"
#include "AppController.h"
#include "Fragment.h"
#include "Stack.h"
#include "PHGeometry.h"
#include "ImageWindow.h"

@interface PictureView (PrivateMethods)

- (void) prepareCoordinates;
- (float) scaleForFragment: (Fragment *) fragm;

- (void) drawRect:(NSRect) rect;

- (BOOL) tryLLJPEG;
- (BOOL) canEnableLLJPEG;

@end // @interface PictureView (PrivateMethods)


@implementation PictureView

- (id) initWithFrame: (NSRect) rect
{
    [super initWithFrame: rect];
    _image = nil;
    _sketch = nil;

    _rotation = 0;

    _fragm = nil;
    _fragms = [[Stack alloc] init];

    _fragmToSketch = nil;

    _selection = nil;

    return self;
}

- (void) dealloc
{
    RELEASE(_image);
    RELEASE(_sketch);

    RELEASE(_fragm);
    RELEASE(_fragms);
    RELEASE(_fragmToSketch);

    RELEASE(_selection);

    [super dealloc];
}

- (void) setImage: (NSImage *) image
{
    // Keep original image
    
    ASSIGN(_image, image);

    ASSIGN(_fragm, [Fragment fragmentWithImage: image
                             //scale: 1.0
                             rotation: _rotation]);

    [_fragms removeAllObjects];

    DESTROY(_selection);

    [self prepareCoordinates];
    [self setNeedsDisplay: YES];
}

- (NSImage *) image
{
    return _image;
}

- (void) clip: (id) sender
{
    if (_selection == nil)
        return;

    // Rotate the fragment to reflect the current appearence of scetch.
    // This will make less surprises on [-undo].
    [_fragm setRotation: _rotation];

    [_fragms push: _fragm ];
    
    Fragment * new_fragm =
        [Fragment fragmentWithScene: [_selection selectedInImage]
                  rotation: _rotation];

    ASSIGN(_fragm, new_fragm);

    DESTROY(_selection);

    [self prepareCoordinates];

    [self setNeedsDisplay: YES];
}

- (void) undo: (id) sender
{
    if ([_fragms depth] > 0)
    {
        Fragment * oldFragment = _fragm;
        [oldFragment retain];
        
        ASSIGN(_fragm, [_fragms pop]);
        _rotation = [_fragm rotation];

        DESTROY(_selection);

        _selection = [[Selection alloc] initWithView: self
                                        selectedInImage: [oldFragment scene]];

        [oldFragment release];

        [self prepareCoordinates];

        [self setNeedsDisplay: YES];
    }
}

- (Fragment *) fragmentToSave
{
    NSRect rect = ( _selection ) ?
        [_selection selectedInImage] : [ _fragm scene ];

    return [Fragment fragmentWithScene: rect rotation : _rotation];
}

- (BOOL) tryLLJPEG
{
    ImageWindow * win = (ImageWindow*)[self window];
    return [win tryLLJPEG];
}

- (BOOL) canEnableLLJPEG
{
    ImageWindow * win = (ImageWindow*)[self window];
    return [win canEnableLLJPEG];
}

- (NSSize) imageSize                   // image coord
{
    return [_image size];
}

- (NSRect) scene                       // image coord
{
    return [_fragm scene];
}

- (NSSize) sketchSize                  // sketch coord
{
    return PHIntegerSize( _sketch ? [_sketch size] : NSZeroSize );
}

- (NSAffineTransform *) fragmToSketch
{
    return _fragmToSketch;
}

- (void) prepareCoordinates
{
    NSRect fragm_scene = _fragm ? [_fragm scene] : NSZeroRect;
    float scale = _fragm ? [self scaleForFragment: _fragm] : 1.0;

    ASSIGN(_fragmToSketch,
           [NSAffineTransform transformToQIofRect: fragm_scene
                              scale: scale
                              rotation: _rotation]);
    

    ASSIGN(_sketch, [_image imageFromRect: fragm_scene
                            scaledBy: scale
                            rotatedBy: _rotation]);

    NSSize frame_size = [self frame].size;
    NSSize sketch_size = _sketch ? [_sketch size] : NSZeroSize ;

    NSPoint sketch_origin = NSMakePoint(
        (frame_size.width - sketch_size.width)/2,
        (frame_size.height - sketch_size.height)/2);

    sketch_origin.x = (int) sketch_origin.x;
    sketch_origin.y = (int) sketch_origin.y;
    
    [self setBoundsOrigin: NSMakePoint(-sketch_origin.x, -sketch_origin.y)];

    [_selection update];
}

- (float) scaleForFragment: (Fragment *) fragm
{
    NSSize rotated_size = [fragm scene].size;
    switch (_rotation)
    {
        case 0:
        case 180:
            break;
        case 90:
        case 270:
        {
            // swap with and height
            float a = rotated_size.width ;
            rotated_size.width = rotated_size.height;
            rotated_size.height = a;
        }
        break;
        default:
        {
            NSAssert1(0, @"Invalid rotation value %d", _rotation);
        }
        break;
    }

    NSSize destination = [self frame].size;

    float magn_x = MIN(destination.width / rotated_size.width, 1);
    float magn_y = MIN(destination.height / rotated_size.height, 1);
    return MIN(magn_x, magn_y);
}

- (void) setRotationTo: (float) degrees
{
    _rotation = degrees;
}

- (void) rotateCCW: (id) sender
{
    [self rotateWithDirection: PHPositiveDirection];
}

- (void) rotateCW: (id) sender
{
    [self rotateWithDirection: PHNegativeDirection];
}

- (void) rotateWithDirection: (PHRotateDirection) direction
{
    int rotation = 0;
    switch (direction)
    {
        case PHPositiveDirection: rotation = 90; break;
        case PHNegativeDirection: rotation = -90; break;
        default:
            NSLog(@"rotateWithDirection: unknown direction %d\n", direction);
            break;
    }

    _rotation += rotation;
    _rotation %= 360;
    if (_rotation < 0)
        _rotation += 360;

    
    [self prepareCoordinates];
    [self setNeedsDisplay: YES];
}

- (void) windowDidResize: (NSNotification *)notification
{
    [self prepareCoordinates];
    [self setNeedsDisplay: YES];
}

- (void) drawRect:(NSRect)frame
{
    if (! _sketch)
        return;

    // draw full sketch ignoring the 'frame' rect
    [_sketch compositeToPoint: NSZeroPoint operation: NSCompositeCopy];

    /*
    // draw the intersection between the sketch and the frame rectangles -
    // off-pixel problem (rounding)?
    
    NSRect sketchRect =
        NSMakeRect(0, 0, [_sketch size].width, [_sketch size].height);
    frame = NSIntersectionRect(frame, sketchRect);
        
    [_sketch compositeToPoint: frame.origin fromRect: frame
             operation: NSCompositeCopy];
    */

    if ( _selection != nil )
    {
        /*
        // draw full mask ignoring the 'frame' rect
        [[_selection mask] compositeToPoint: NSZeroPoint
                           operation: NSCompositeSourceOver];
        */

        NSRect maskRect = NSInsetRect(frame, -1, -1);
        NSImage * mask = [_selection maskForRect: maskRect];

        
        [mask compositeToPoint: maskRect.origin
              operation: NSCompositeSourceOver];
    }
}

- (void) resetCursorRects
{
    if ( _selection )
    {
        NSArray * rects = [_selection cursorRects];

        [rects makeObjectsPerformSelector: @selector(addToView:)
               withObject: self];

        [self addCursorRect: [self frame] cursor: [NSCursor arrowCursor]];
    }
}

- (void) mouseDown:(NSEvent *) event
{
    NSPoint pt = [self convertPoint: [event locationInWindow] fromView: nil];

    [self discardCursorRects];

    if ( _selection == nil )
        _selection = [[Selection alloc] initWithView: self];
                                                                     
    [_selection mouseDown: pt];

    [[_selection trackingCursor] set];

    [self setNeedsDisplayInRect: [_selection updatedInSketch]];
}

- (void) mouseUp:(NSEvent *) event 
{
    if (_selection != nil)
    {
        NSPoint pt = [self convertPoint:[event locationInWindow] fromView:nil];
        
        [_selection mouseUp: pt];
        if ( ! [_selection valid] )
          DESTROY( _selection );

        if ( _selection && [self tryLLJPEG] )
            [_selection adjustToJPEGGrid: [_image jpegBlockSize]];
        
        [self setNeedsDisplayInRect:
                (_selection ? [_selection updatedInSketch] : [self frame]) ];
    }

    [[NSCursor arrowCursor] set];   // restore default cursor
    
    // invalidateCursorRectsForView does not work here.
    // something about self not being key (?)
    //[[self window] invalidateCursorRectsForView: self];
    [[self window] resetCursorRects];
}
 
- (void) mouseDragged:(NSEvent *) event
{
    if (_selection != nil)
    {
        NSPoint pt = [self convertPoint:[event locationInWindow] fromView:nil];
        
        [_selection mouseDragged: pt];

        [self setNeedsDisplayInRect: [_selection updatedInSketch]];
    }
}

- (void) toggleLLJPEG: (id) sender
{
  BOOL currentState = [[[NSApp delegate] selectionCtrl] tryLLJPEG];
  [[[NSApp delegate] selectionCtrl] setTryLLJPEG: ! currentState];

  if ( _selection && [self tryLLJPEG] )
  {
      [_selection adjustToJPEGGrid: [_image jpegBlockSize]];
      [self setNeedsDisplayInRect: [_selection updatedInSketch]];
  }
  
}

- (BOOL) acceptsFirstResponder
{
    return YES;
}

- (void) keyDown:(NSEvent *) event
{
    // print the event
    //if ( [ event characters ] )
    //    NSLog( @"keyDown: %@\n", [ event characters ] );

    // If this is a keyboard action send the message bound to it along
    // responder chain.
    [self interpretKeyEvents: [NSArray arrayWithObject: event]];
}

- (BOOL) validateMenuItem: (NSMenuItem *) item
{
  SEL action = [item action];

  if ( sel_eq(action, @selector(rotateCW:)) ||
       sel_eq(action, @selector(rotateCCW:)) )
  {
    return _image ? YES : NO;
  }

  if ( sel_eq(action, @selector(clip:)) )
  {
    return _selection ? YES : NO;
  }

  if ( sel_eq(action, @selector(undo:)) )
  {
    return ([_fragms empty] == NO) ? YES : NO;
  }

  if ( sel_eq(action, @selector(toggleLLJPEG:) ) )
  {
    BOOL canEnable = [self canEnableLLJPEG];
    BOOL isOn = [[[NSApp delegate] selectionCtrl] tryLLJPEG];
    [item setState: (isOn && canEnable) ? NSOnState : NSOffState];
    return canEnable;
  }
  
  return YES;
}

@end
