#include "Selection.h"
#include <AppKit/NSAffineTransform.h>
#include <AppKit/NSEvent.h>
#include <AppKit/NSColor.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSApplication.h>
#include <AppKit/NSCursor.h>
#include "PHGeometry.h"
#include "PHAffineTransform.h"
#include "AppController.h"
#include "SelectionCtrl.h"
#include "MouseTracker.h"
#include "CursorRectInfo.h"
#include <math.h>

typedef enum { PHOutsideRect = 0, PHInsideRect, PHCornerRect, PHEdgeRect }
PHRectRegion;

const static int BW = 2; // boundary width / 2

@interface Selection (PrivateMethods)

- (NSRect) sketch;
- (BOOL) correctSelection;

- (MouseTracker *) createMouseTrackerWithStartPoint: (NSPoint) pt;
- (int) zoneOfRect: (NSRect) rc fromPoint: (NSPoint) pt;

@end // Selection (PrivateMethods)


@implementation Selection

- (id) initWithView: (PictureView *) view
{
    return [self initWithView: view selectedInImage: NSZeroRect];
}

- (id) initWithView: (PictureView *) view selectedInImage: (NSRect) rc
{
    _view = view;
    _selectedInImage = rc;

    [self update];
    
    _previousSelection = NSZeroRect;
    
    _tracker = nil;
    return self;
}

- (void) update
{
    _selection = [[_view fragmToSketch] transformRect: _selectedInImage];
    _selection = PHIntegerRect(_selection);

    // After the transformation the selection in image may be bigger
    // than the sketch because the sketch was truncated to integer size
    // (see [NSImage(PhotoClip) -imageFromRect:scaledBy:rotatedBy:].

    _selection = NSIntersectionRect( _selection, [self sketch] );
}

- (void) dealloc
{
    RELEASE(_tracker);
    [super dealloc];
}

- (void) mouseDown: (NSPoint) pt
{
    ASSIGN(_tracker, [self createMouseTrackerWithStartPoint: pt]);

    _selection = [_tracker selectionRect];

    // force full redraw if selection is invalid
    _previousSelection = [self valid] ? _selection : [self sketch];
}

- (void) mouseDragged: (NSPoint) pt
{
    NSAssert(_tracker, @"_tracker == nil");

    //NSLog(@"Selection: %@", NSStringFromPoint(pt) );
    
    _previousSelection = [_tracker selectionRect];
    _selection = [_tracker mouseDragged: pt];
}

- (void) mouseUp: (NSPoint) pt
{
    // I have to comment it because of low-pass filter: it adds "inertia"
    // No, on slow displays (over network) this seems compulsory
    [self mouseDragged: pt];

    if ([self valid]) // if nonempty
        [self correctSelection];
}

- (BOOL) correctSelection
{
    NSAffineTransform * reverse = PHReverseTransform( [_view fragmToSketch] );

    // _selection is in sketch coordinates, result is in image coordinates
    NSRect result = [reverse transformRect: _selection];

    result = NSIntersectionRect( [_view scene], result );

    // Round coordinates to integer numbers.

    float requiredAR = [[[NSApp delegate] selectionCtrl] shape];
    result = PHIntegerRectPreserveAR( result, requiredAR );
    BOOL isAdjusted = NO;

    // If we just dragged the selection, do not resize
    if ( [_tracker class] ==  [MouseTrackerMoveAll class] )
      _selectedInImage.origin = result.origin;
    else
      _selectedInImage = result;
    
    
    // Stretch the selected rect to the required size, if nesessary

    int requiredSpan = [[[NSApp delegate] selectionCtrl] span];
    int currentSpan = PHMaxSpan( _selectedInImage.size );
    if ( 0 < requiredSpan && currentSpan < requiredSpan )
    {
        NSRect result = PHStretchRectToSpan( _selectedInImage, requiredSpan,
                                             requiredAR, [_view scene] );

        isAdjusted = YES;
        if ( NSIsEmptyRect( result ) )
        {
            // TODO: indicate to the user that we cannot stretch
            isAdjusted = NO;
        }
        else if ( NSEqualRects( result, [_view scene] ) )
        {
            // Let the parent view destroy the selection if the whole
            // scene became selected as a result of this stretch.
            // TODO: indicate to the user what's going on.
            result = NSZeroRect;
        }
        
        if ( isAdjusted )
        {
          _selectedInImage = result;
          [self update];
        }
    }

    return isAdjusted;
}

- (void) adjustToJPEGGrid: (NSSize) gridSize
{
    int requiredSpan = [[[NSApp delegate] selectionCtrl] span];
    BOOL isSpanRequired = ([[[NSApp delegate] selectionCtrl] span] > 0);
    BOOL isARRequired = ([[[NSApp delegate] selectionCtrl] shape] > 0);

    // Adjust to grid only if we do not scale
    if ( isSpanRequired && PHMaxSpan(_selectedInImage.size) > requiredSpan )
        return; // we will scale the fragment down

    NSRect result = _selectedInImage;

    // The following adjustments should move the rectangle in
    // the direction opposite to PHIntegerRectPreserveAR()
    // in the method [-correctSelection], otherwise we get into loop
    // and each mouse click in the middle of selected rect moves it
    // to the next block boundary.
    
    if ( isSpanRequired )
    {
        // Since we do not scale the fragment, the current span
        // is equial to the required one and we need to preserve it
        if ( isARRequired )
            // move origin and hope for the best
            result = PHJMoveRectOrigin(result, gridSize, [_view imageSize]);
        else
            // keep bigger and inflate smaller side
            result = PHJMoveRectKeepSpan(result, gridSize, [_view imageSize]);
    }
    else
    {
        if ( isARRequired )
            // inflate bigger size, set smaller size as close to AR as possible
            result = PHJMoveRectKeepAR(result, gridSize, [_view imageSize]);
        else
            // the same
            result = PHJMoveRectKeepAR(result, gridSize, [_view imageSize]);
    }

    //NSLog( @"JPEG adjusted: %@", PHStringFromRect( result ) );

    _selectedInImage = result;
    [self update];
}

- (MouseTracker *) createMouseTrackerWithStartPoint: (NSPoint) pt
{
    MouseTracker * tracker = nil;

    int area = [self zoneOfRect: _selection fromPoint: pt];

    Class cls;
    switch (area)
    {
        case PHInsideRect:
            cls = [MouseTrackerMoveAll class];
            break;
        case PHEdgeRect:
            cls = [MouseTrackerMoveCorner class];
            break;
        case PHOutsideRect:
            cls = [MouseTrackerNewSel class];
            break;
        default:
            NSAssert1(0, @"unknown rectangle area=%d", area);
            cls = [MouseTrackerNewSel class];
            break;
    }
    
    tracker = [[cls alloc] initWithFrame: [self sketch]
                           selRect: _selection
                           start: pt];

    return AUTORELEASE(tracker);
}

- (int) zoneOfRect: (NSRect) rc fromPoint: (NSPoint) pt
{
    
    NSRect left = NSMakeRect(NSMinX(rc) - BW, NSMinY(rc) - BW,
                             2*BW, 2*BW + rc.size.height);
    NSRect right = NSMakeRect(NSMaxX(rc) - BW, NSMinY(rc) - BW,
                              2*BW, 2*BW + rc.size.height);
    NSRect bottom = NSMakeRect(NSMinX(rc) - BW, NSMinY(rc) - BW,
                               2*BW + rc.size.width, 2*BW);
    NSRect top = NSMakeRect(NSMinX(rc) - BW, NSMaxY(rc) - BW,
                            2*BW + rc.size.width, 2*BW);

    BOOL on_edge =
        NSPointInRect(pt, left) || NSPointInRect(pt, right) ||
        NSPointInRect(pt, bottom) || NSPointInRect(pt, top);

    if (on_edge)
        return PHEdgeRect;
    else if (NSPointInRect(pt, rc))
        return PHInsideRect;
    else
        return PHOutsideRect;
        
}

- (NSArray *) cursorRects
{
    NSMutableArray * result = [NSMutableArray arrayWithCapacity: 2];
    NSRect rc;

    NSCursor * inner = [MouseTrackerMoveAll cursorType];
    NSCursor * edge = [MouseTrackerNewSel cursorType];
    
    rc = NSInsetRect(_selection, -BW, -BW);
    [result addObject: [CursorRectInfo infoWithRect: rc cursor: edge]];

    rc = NSInsetRect(_selection, BW, BW);
    [result addObject: [CursorRectInfo infoWithRect: rc cursor: inner]];

    return result;
}

- (BOOL) isOutOfSelPoint: (NSPoint) pt
{
    return
        ! [self valid] ||
        ! NSPointInRect(pt, NSInsetRect(_selection, -BW, -BW));
}

- (NSCursor *) trackingCursor
{
    return (_tracker) ? [_tracker cursor] : NULL;
}

- (BOOL) valid
{
    return NSIsEmptyRect([self selectedInSketch]) == NO;
}

- (NSRect) sketch
{
    NSRect rc;
    rc.origin = NSZeroPoint;
    rc.size = [_view sketchSize];
    return rc;
}
- (NSRect) selectedInImage
{
    return _selectedInImage;
}

- (NSRect) selectedInSketch
{
    return NSIntersectionRect([self sketch], _selection);

    // _selection should never extend beyond _sketch, so can just
    //return _selection;
}

- (NSRect) updatedInSketch
{
    return NSInsetRect(
        NSUnionRect(_previousSelection, [self selectedInSketch]),
        -1, -1);
}

- (NSImage *) mask
{
    NSImage * mask = [[NSImage alloc] initWithSize: [_view sketchSize]];
    
    NSColor * blur = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.57];
    NSColor * transp = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.0];

    [mask lockFocus];
    
    [blur set];

    NSRectFillUsingOperation([self sketch], NSCompositeCopy);

    [transp set];

    NSRect selectionRect = [self selectedInSketch];
    
    NSRectFillUsingOperation(selectionRect, NSCompositeCopy);
    
    [mask unlockFocus];

    return AUTORELEASE(mask);
}

- (NSImage *) maskForRect: (NSRect) rect
{
    NSImage * mask = [[NSImage alloc] initWithSize: rect.size];

    NSColor * blur = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.57];
    NSColor * transp = [NSColor colorWithCalibratedWhite: 0.0 alpha: 0.0];

    [mask lockFocus];
    
    [blur set];

    NSRect maskRect = NSMakeRect(0, 0, rect.size.width, rect.size.height);

    // sketch rect in mask coordinates
    NSRect imageRect = NSOffsetRect([self sketch],
                                    -rect.origin.x, -rect.origin.y);
    
    NSRectFillUsingOperation(NSIntersectionRect(maskRect, imageRect),
                             NSCompositeCopy);

    [transp set];

    // in sketch coordinates
    NSRect selectionRect = [self selectedInSketch];

    // convert to mask coordinates
    NSRect maskSelRect = NSOffsetRect(selectionRect,
                                      -rect.origin.x, -rect.origin.y);

    maskSelRect = NSIntersectionRect(maskSelRect, maskRect);
    
    NSRectFillUsingOperation(NSIntersectionRect(maskSelRect, imageRect),
                             NSCompositeCopy);
    
    [mask unlockFocus];

    return AUTORELEASE(mask);
}

@end // implementation Selection
